using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

namespace EmissiveLighting
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        Effect emissiveEffect;
        Model model;
        Matrix[] modelTransforms;

        Matrix world;
        Matrix view;
        Matrix projection;

        // This is both the light color and the light intensity 
        Vector3 ambientLightColor;
        // The color of the objects you are going to draw
        Vector3[] diffuseColor;
        // The emissive color of the objects
        Vector3[] emissiveColor;
        // The direction the light comes from
        Vector3 lightDirection;
        // The color and intensity of the diffuse light
        Vector3 diffuseLightColor;
        // The direction and color of a 2nd light
        Vector3 lightDirection2;
        Vector3 diffuseLightColor2;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // Set the values for our world, view, and projection matrices
            world = Matrix.Identity;
            view = Matrix.CreateLookAt(new Vector3(0, 1.5f, 3.5f), new Vector3(0, 0, 0), new Vector3(0, 1, 0));
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, GraphicsDevice.Viewport.AspectRatio, 0.1f, 100.0f);

            // Set the ambient color
            ambientLightColor = new Vector3(0.05f, 0.1f, 0.2f);

            // Set the diffuse colors
            diffuseColor = new Vector3[4];
            diffuseColor[0] = new Vector3(1, 0.25f, 0.25f);
            diffuseColor[1] = new Vector3(0.25f, 1, 0.25f);
            diffuseColor[2] = new Vector3(0.25f, 0.25f, 1);
            diffuseColor[3] = new Vector3(0.5f, 0.0f, 0.5f);

            // Set the emissive colors
            emissiveColor = new Vector3[4];
            emissiveColor[0] = new Vector3(0, 0.75f, 0);
            emissiveColor[1] = new Vector3(0, 0, 0.75f);
            emissiveColor[2] = new Vector3(0.75f, 0, 0);
            emissiveColor[3] = new Vector3(0, 0.75f, 0);

            // Set light starting location
            lightDirection = new Vector3(-0.5f, -0.5f, -0.6f);
            lightDirection.Normalize();

            // Set the lights diffuse color
            diffuseLightColor = new Vector3(1, 0.9f, 0.8f);

            // Set the 2nd lights direction and color
            lightDirection2 = new Vector3(0.45f, -0.8f, 0.45f);
            lightDirection2.Normalize();
            diffuseLightColor2 = new Vector3(0.4f, 0.35f, 0.4f);

            BasicEffect test = new BasicEffect(GraphicsDevice);
            test.EnableDefaultLighting();

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // Load custom effect
            emissiveEffect = Content.Load<Effect>("EmissiveLighting");

            // Load our model
            model = Content.Load<Model>("LightingModels");

            // Create an array for model transforms and make a copy
            modelTransforms = new Matrix[model.Bones.Count];
            model.CopyAbsoluteBoneTransformsTo(modelTransforms);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // TODO: Add your update logic here

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            int colorIndex = 0;

            // Rotate and normalize the light
            Matrix rotate = Matrix.CreateRotationY((float)gameTime.ElapsedGameTime.TotalSeconds) *
                            Matrix.CreateRotationZ((float)gameTime.ElapsedGameTime.TotalSeconds);
            lightDirection = Vector3.TransformNormal(lightDirection, rotate);
            lightDirection.Normalize();

            // Set effect wide parameters
            emissiveEffect.Parameters["View"].SetValue(view);
            emissiveEffect.Parameters["Projection"].SetValue(projection);
            emissiveEffect.Parameters["AmbientLightColor"].SetValue(ambientLightColor);
            emissiveEffect.Parameters["LightDirection"].SetValue(-lightDirection);
            emissiveEffect.Parameters["DiffuseLightColor"].SetValue(diffuseLightColor);
            emissiveEffect.Parameters["LightDirection2"].SetValue(-lightDirection2);
            emissiveEffect.Parameters["DiffuseLightColor2"].SetValue(diffuseLightColor2);

            foreach (ModelMesh mesh in model.Meshes)
            {
                // Set mesh effect parameters
                emissiveEffect.Parameters["World"].SetValue(modelTransforms[mesh.ParentBone.Index] * world);

                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                {
                    // Set vertex and index buffer to use
                    GraphicsDevice.SetVertexBuffer(meshPart.VertexBuffer, meshPart.VertexOffset);
                    GraphicsDevice.Indices = meshPart.IndexBuffer;

                    // Set diffuse color for the object
                    emissiveEffect.Parameters["DiffuseColor"].SetValue(diffuseColor[colorIndex]);
                    emissiveEffect.Parameters["EmissiveColor"].SetValue(emissiveColor[colorIndex]);
                    colorIndex++;

                    // Apply our one and only pass
                    emissiveEffect.CurrentTechnique.Passes[0].Apply();

                    GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0,
                                                         meshPart.NumVertices,
                                                         meshPart.StartIndex,
                                                         meshPart.PrimitiveCount);
                }
            }


            base.Draw(gameTime);
        }
    }
}
