using System;
using System.IO;
using System.ComponentModel;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Serialization.Compiler;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;

namespace ContentPipelineExtension
{
    public class MyLevelObject
    {
        public ModelContent Model;
        public Vector3 Position;
        public Vector3 Rotation;
        public string ModelFileName;
    }

    public class Level
    {
        public List<MyLevelObject> LevelObjects;
        public TextureContent Background;
        public string BackgroundFileName;
    }

    /// <summary>
    /// This class will be instantiated by the XNA Framework Content Pipeline
    /// to apply custom processing to content data, converting an object of
    /// type TInput to TOutput. The input and output types may be the same if
    /// the processor wishes to alter data without changing its type.
    ///
    /// This should be part of a Content Pipeline Extension Library project.
    ///
    /// TODO: change the ContentProcessor attribute to specify the correct
    /// display name for this processor.
    /// </summary>
    [ContentImporter(".mylevel", DisplayName = "My Level Importer", 
        DefaultProcessor = "MyLevelProcessor")]
    public class MyContentImporter : ContentImporter<Level>
    {
        public override Level Import(string filename, 
            ContentImporterContext context)
        {
            Level level = new Level();
            level.LevelObjects = new List<MyLevelObject>();
            using (StreamReader reader = new StreamReader(File.OpenRead(filename)))
            {
                // The first line is the background texture
                level.BackgroundFileName = GetAbsolutePath(filename, reader.ReadLine());
                while (!reader.EndOfStream)
                {
                    // After the first line, each line is first the model, 
                    // and then the model's data
                    MyLevelObject info = new MyLevelObject();
                    info.ModelFileName = GetAbsolutePath(filename, reader.ReadLine());
                    // Followed by the position/rotation data
                    string[] floats = reader.ReadLine().Split(' ');
                    info.Position = new Vector3(float.Parse(floats[0]),
                        float.Parse(floats[1]), float.Parse(floats[2]));
                    info.Rotation = new Vector3(float.Parse(floats[3]),
                        float.Parse(floats[4]), float.Parse(floats[5]));
                    
                    level.LevelObjects.Add(info);
                }
            }

            return level;
        }

        private string GetAbsolutePath(string input, string file)
        {
            return Path.Combine(Path.GetDirectoryName(input), file);
        }
    }

    [ContentProcessor(DisplayName="MyLevelProcessor")]
    public class MyContentProcessor : ContentProcessor<Level, Level>
    {
        public override Level Process(Level input, ContentProcessorContext context)
        {
            Level level = new Level();
            level.LevelObjects = new List<MyLevelObject>();
            OpaqueDataDictionary opaqueData = new OpaqueDataDictionary();
            string textureProcessor = ShouldTexturesBeInverted ?
                "InvertColorsProcessor" : "TextureProcessor";
            if (ShouldTexturesBeInverted)
            {
                opaqueData.Add("ShouldCauseBlocks", ShouldCauseBlocks);
                opaqueData.Add("BlockSize", BlockSize);
            }

            // Process the input, we'll need to create the texture that is specified
            ExternalReference<TextureContent> backgroundReference = 
                new ExternalReference<TextureContent>(input.BackgroundFileName);
            
            // Now build and store it
            level.Background = context.BuildAndLoadAsset<TextureContent,
                TextureContent>(backgroundReference, textureProcessor, opaqueData, null);
            // Now load each of the models
            foreach (MyLevelObject obj in input.LevelObjects)
            {
                MyLevelObject newobj = new MyLevelObject();
                newobj.Position = obj.Position;
                newobj.Rotation = obj.Rotation;

                // Do the same thing we did for textures for model here
                ExternalReference<NodeContent> modelReference = 
                    new ExternalReference<NodeContent>(obj.ModelFileName);
                // Now build and store it
                newobj.Model = context.BuildAndLoadAsset<NodeContent, 
                    ModelContent>(modelReference, "ModelProcessor");
                level.LevelObjects.Add(newobj);
            }

            return level;
        }

        [DefaultValue(false)]
        [DisplayName("ShouldInvertColors")]
        [Description("Should the background texture be inverted?")]
        public bool ShouldTexturesBeInverted
        {
            get;
            set;
        }

        [DefaultValue(false)]
        [DisplayName("BlockyInversion")]
        [Description("Should the inversion be done in blocks?")]
        public bool ShouldCauseBlocks
        {
            get;
            set;
        }

        [DefaultValue(20)]
        [DisplayName("BlockSize")]
        [Description("The size in pixels of the blocks if a blocky inversion is done")]
        public int BlockSize
        {
            get;
            set;
        }
    }

    [ContentTypeWriter]
    public class LevelContentWriter : ContentTypeWriter<Level>
    {
        public override string GetRuntimeReader(TargetPlatform targetPlatform)
        {
            return "ContentImporter.LevelContentReader, ContentImporter";
        }
        public override string GetRuntimeType(TargetPlatform targetPlatform)
        {
            return "ContentImporter.Level, ContentImporter";
        }
        protected override void Write(ContentWriter output, Level value)
        {
            // Write the background first
            output.WriteObject<TextureContent>(value.Background);
            // Next write the count, so we know how many models there are
            output.Write(value.LevelObjects.Count);
            // Finally, write all the level objects
            foreach (MyLevelObject obj in value.LevelObjects)
            {
                output.WriteObject<ModelContent>(obj.Model);
                output.Write(obj.Position);
                output.Write(obj.Rotation);
            }
        }
    }

    [ContentProcessor(DisplayName = "InvertColorsProcessor")]
    public class InvertColorsProcessor : ContentProcessor<TextureContent, TextureContent>
    {
        public override TextureContent Process(TextureContent input,
            ContentProcessorContext context)
        {
            // Convert the input to standard Color format, for ease of processing.            
            input.ConvertBitmapType(typeof(PixelBitmapContent<Color>));
            foreach (MipmapChain imageFace in input.Faces)
            {
                for (int i = 0; i < imageFace.Count; i++)
                {
                    PixelBitmapContent<Color> mip = (PixelBitmapContent<Color>)imageFace[i];
                    // Invert the colors
                    for (int w = 0; w < mip.Width; w++)
                    {
                        for (int h = 0; h < mip.Height; h++)
                        {
                            Color original = mip.GetPixel(w, h);
                            Color inverted = new Color(255 - original.R,
                                255 - original.G, 255 - original.B);
                            if (ShouldCauseBlocks)
                            {
                                if ((h % BlockSize > (BlockSize / 2))
                                    || (w % BlockSize > (BlockSize / 2)))
                                {
                                    inverted = original;
                                }
                            }
                            mip.SetPixel(w, h, inverted);

                        }
                    }
                }
            }
            return input;
        }

        [DefaultValue(false)]
        [DisplayName("BlockyInversion")]
        [Description("Should the inversion be done in blocks?")]
        public bool ShouldCauseBlocks
        {
            get;
            set;
        }

        [DefaultValue(20)]
        [DisplayName("BlockSize")]
        [Description("The size in pixels of the blocks if a blocky inversion is done")]
        public int BlockSize
        {
            get;
            set;
        }
    }
}