using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.GamerServices;

namespace AvatarAnimationBlending
{
    public class AnimationBlender : IAvatarAnimation
    {
        // The current aniamtion that is playing
        private AvatarAnimation currentAnimation;
        // The animation to blend to
        private AvatarAnimation nextAnimation;
        // The total time to blend to the next animation
        private TimeSpan totalBlendTime = TimeSpan.FromMilliseconds(300);
        // How far along the blend we are
        private TimeSpan currentBlendTime;

        /// <summary>
        /// Returns the bone transforms to use with the AvatarRenderer
        /// </summary>
        public ReadOnlyCollection<Matrix> BoneTransforms
        {
            get { return boneTransforms; }
        }
        private Matrix[] blendedAvatarBones = new Matrix[AvatarRenderer.BoneCount];
        private ReadOnlyCollection<Matrix> boneTransforms;

        /// <summary>
        /// The current position in the animation
        /// If we are blending then we use the next animation
        /// </summary>
        public TimeSpan CurrentPosition
        {
            get
            {
                TimeSpan returnValue;

                if (nextAnimation != null)
                    returnValue = nextAnimation.CurrentPosition;
                else
                    returnValue = currentAnimation.CurrentPosition;

                // Call update to set the avatar transforms
                Update(TimeSpan.Zero, false);

                return returnValue;
            }
            set
            {
                if (nextAnimation != null)
                    nextAnimation.CurrentPosition = value;
                else
                    currentAnimation.CurrentPosition = value;
            }
        }

        /// <summary>
        /// The current expression in the animation
        /// If we are blending then we use the next animation
        /// The expression is not blended so we just use 
        /// the next animations expression
        /// </summary>
        public AvatarExpression Expression
        {
            get 
            {
                if (nextAnimation != null)
                    return nextAnimation.Expression;
                else
                    return currentAnimation.Expression;
            }
        }

        /// <summary>
        /// The length of the animation
        /// If we are blending then we use the next animation
        /// </summary>
        public TimeSpan Length
        {
            get
            {
                if (nextAnimation != null)
                    return nextAnimation.Length;
                else
                    return currentAnimation.Length;
            }
        }

        public AnimationBlender(AvatarAnimation startingAnimation)
        {
            currentAnimation = startingAnimation;
            
            boneTransforms = new ReadOnlyCollection<Matrix>(blendedAvatarBones);
        }

        /// <summary>
        /// Updates the animation and performs blending to the next animation
        /// </summary>
        public void Update(TimeSpan elapsedAnimationTime, bool loop)
        {
            // Update the current animation
            currentAnimation.Update(elapsedAnimationTime, loop);

            // If we are not blending to an animation just copy the current transforms
            if (nextAnimation == null)
            {
                currentAnimation.BoneTransforms.CopyTo(blendedAvatarBones, 0);
                return;
            }

            // If we get here we are blending and we need to update the next animation
            nextAnimation.Update(elapsedAnimationTime, loop);

            // Update where we are in the blend
            currentBlendTime += elapsedAnimationTime;

            // Calculate blend factor
            float blendAmount = (float)(currentBlendTime.TotalSeconds /
                                        totalBlendTime.TotalSeconds);


            // If the blend is over 1 then we are done blending
            // We can just use the next animation transforms
            if (blendAmount >= 1.0f)
            {
                // Set the current animtion and remove the target animation
                currentAnimation = nextAnimation;
                nextAnimation.BoneTransforms.CopyTo(blendedAvatarBones, 0);
                nextAnimation = null;
                return;
            }

            // We need to store the rotation and translation for each bone in 
            // the current and next animations. We then need to calculate the 
            // final values using the blend amount as a weight between the 
            // current and next values.
            Quaternion currentRotation, nextRotation, blendedRotation;
            Vector3 currentTranslation, nextTranslation, blendedTranslation;

            // Loop all bone transforms
            for (int i = 0; i < blendedAvatarBones.Length; ++i)
            {
                currentRotation = Quaternion.CreateFromRotationMatrix(currentAnimation.BoneTransforms[i]);
                currentTranslation = currentAnimation.BoneTransforms[i].Translation;
                nextRotation = Quaternion.CreateFromRotationMatrix(nextAnimation.BoneTransforms[i]);
                nextTranslation = nextAnimation.BoneTransforms[i].Translation;

                // Calculate the blended rotation
                Quaternion.Slerp(ref currentRotation, ref nextRotation, blendAmount, out blendedRotation);
                
                // Calculate the blended translation
                Vector3.Lerp(ref currentTranslation, ref nextTranslation, blendAmount, out blendedTranslation);

                // Create the blended bone matrix
                blendedAvatarBones[i] = Matrix.CreateFromQuaternion(blendedRotation) * Matrix.CreateTranslation(blendedTranslation);
            }
        }

        /// <summary>
        /// Starts the process of blending to the next animation
        /// </summary>
        public void Play(AvatarAnimation animation)
        {
            // Animation to blend to
            nextAnimation = animation;
            // Reset the animation to start at the beginning
            nextAnimation.CurrentPosition = TimeSpan.Zero;

            // Set our current blend position
            currentBlendTime = TimeSpan.Zero;
        }
    }
}
