using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;
using Microsoft.Xna.Framework.GamerServices;
using CustomAvatarAnimation;

namespace CustomAvatarAnimationPipelineExtension
{
    [ContentProcessor(DisplayName = "CustomAvatarAnimationProcessor")]
    public class CustomAvatarAnimationProcessor :
                                 ContentProcessor<NodeContent, CustomAvatarAnimationData>
    {
        // The bind pose of the avatar
        List<Matrix> bindPose = new List<Matrix>();

        public override CustomAvatarAnimationData Process(NodeContent input, ContentProcessorContext context)
        {
            // Find the skeleton of the model
            NodeContent skeleton = FindSkeleton(input);

            // We have to find the skeleton and it needs to have 1 animtion
            if (skeleton == null || skeleton.Animations.Count != 1)
            {
                throw new InvalidContentException("Invalid avatar animation.");
            }

            // Update the skeleton to what we expect at runtime
            CleanSkeleton(skeleton);

            // Flat list of the bones in the skeleton
            IList<NodeContent> bones = FlattenSkeleton(skeleton);

            // The number of bones should match what the AvatarRender expects
            if (bones.Count != AvatarRenderer.BoneCount)
            {
                throw new InvalidContentException("Invalid number of bones found.");
            }

            // Populate the bind pose list
            foreach (NodeContent bone in bones)
            {
                bindPose.Add(bone.Transform);
            }

            // Build up a table mapping bone names to indices
            Dictionary<string, int> boneNameMap = new Dictionary<string, int>();
            for (int i = 0; i < bones.Count; i++)
            {
                string boneName = bones[i].Name;
                if (!string.IsNullOrEmpty(boneName))
                {
                    boneNameMap.Add(boneName, i);
                }
            }

            CustomAvatarAnimationData avatarCustomAnimationData = null;
            foreach (KeyValuePair<string, AnimationContent> animation in skeleton.Animations)
            {
                // Animation duration needs to be greater than 0 length
                if (animation.Value.Duration <= TimeSpan.Zero)
                {
                    throw new InvalidContentException("Animation has a zero duration.");
                }

                // Build a list of the avatar keyframes in the animation
                List<Keyframe> animationKeyFrames = ProcessAnimation(animation.Value, boneNameMap);

                // Check for an invalid keyframes list
                if (animationKeyFrames.Count <= 0)
                {
                    throw new InvalidContentException("Animation has no keyframes.");
                }

                // Create the custom-animation object
                avatarCustomAnimationData = new CustomAvatarAnimationData(animation.Key, animation.Value.Duration, animationKeyFrames);
            }

            return avatarCustomAnimationData;
        }

        private NodeContent FindSkeleton(NodeContent input)
        {
            // This is the node we are looking for
            if (input.Name.Contains("BASE__Skeleton"))
            {
                return input;
            }

            // Recursively check all children until we find the root of the skeleton
            foreach (NodeContent child in input.Children)
            {
                NodeContent skeleton = FindSkeleton(child);

                if (skeleton != null)
                    return skeleton;
            }

            return null;
        }

        // Flatten the skeleton into a list ordered by level depth
        static IList<NodeContent> FlattenSkeleton(NodeContent skeleton)
        {
            // Return list of bones we find in the skeleton
            List<NodeContent> bones = new List<NodeContent>();

            // Skeleton bones in the current level
            List<NodeContent> currentLevelBones = new List<NodeContent>();

            // Start with the root node
            currentLevelBones.Add(skeleton);

            while (currentLevelBones.Count > 0)
            {
                List<NodeContent> nextLevelBones = new List<NodeContent>();

                // Avatar bones are sorted by name in each level
                IEnumerable<NodeContent> sortedBones = from item in currentLevelBones
                                                       orderby item.Name
                                                       select item;

                // Add the sorted list to our list
                foreach (NodeContent bone in sortedBones)
                {
                    bones.Add(bone);

                    // Add all of the children for the next level
                    foreach (NodeContent child in bone.Children)
                    {
                        nextLevelBones.Add(child);
                    }
                }

                currentLevelBones = nextLevelBones;
            }

            return bones;
        }


        // Removes bones not used in the AvatarRenderer at runtime
        // and fixes the names of some bones
        static void CleanSkeleton(NodeContent bone)
        {
            // Remove unwated text from the bone name
            bone.Name = bone.Name.Replace("__Skeleton", "");

            // Process all of the children
            for (int i = 0; i < bone.Children.Count; ++i)
            {
                NodeContent child = bone.Children[i];
                if (child.Name.Contains("_END"))
                {
                    bone.Children.Remove(child);
                    --i;
                }
                else
                {
                    CleanSkeleton(child);
                }
            }
        }


        // Convert animation from content pipeline format to a list of keyframes 
        List<Keyframe> ProcessAnimation(AnimationContent animation, Dictionary<string, int> boneMap)
        {
            // Return keyframe list
            List<Keyframe> keyframes = new List<Keyframe>();

            foreach (KeyValuePair<string, AnimationChannel> channel in animation.Channels)
            {
                // Don't process the end bone channel. We have removed these from the skeleton
                if (channel.Key.Contains("_END"))
                    continue;

                // Find which bone this channel has keyframes for
                int boneIndex;
                if (!boneMap.TryGetValue(channel.Key.Replace("__Skeleton", ""), out boneIndex))
                {
                    throw new InvalidContentException(string.Format(
                        "Found animation for bone '{0}', " +
                        "which is not part of the skeleton.", channel.Key));
                }

                // Craete the keyframes for the channel
                foreach (AnimationKeyframe keyframe in channel.Value)
                {
                    keyframes.Add(new Keyframe(boneIndex, keyframe.Time, CreateKeyframeMatrix(keyframe, boneIndex)));
                }
            }

            // Sort the final list of keyframes by time
            keyframes.Sort((frame1, frame2) => frame1.Time.CompareTo(frame2.Time));

            return keyframes;
        }


        // Convert animation keyframes info the format used by the AvatarRenderer
        Matrix CreateKeyframeMatrix(AnimationKeyframe keyframe, int boneIndex)
        {
            Matrix keyframeMatrix;

            // The root node is transformed by the root of the bind pose
            // We need to make the keyframe relative to the root
            if (boneIndex == 0)
            {
                // If you are using an older verion of the FBX exporter the root 
                // of the bind pose my be translated incorrectly
                // If your model appears to be floating use the following translation
                //Vector3 bindPoseTranslation = new Vector3(0.000f, 75.5199f, -0.8664f);
                Vector3 bindPoseTranslation = Vector3.Zero;

                Matrix inverseBindPose = bindPose[boneIndex];
                inverseBindPose.Translation -= bindPoseTranslation;
                inverseBindPose = Matrix.Invert(inverseBindPose);

                Matrix keyTransfrom = keyframe.Transform;
                keyframeMatrix = (keyTransfrom * inverseBindPose);
                keyframeMatrix.Translation -= bindPoseTranslation;

                // Scale from cm to meters
                keyframeMatrix.Translation *= 0.01f;
            }
            else
            {
                keyframeMatrix = keyframe.Transform;
                // Remove translation from anything by the root
                keyframeMatrix.Translation = Vector3.Zero;
            }

            return keyframeMatrix;
        }
    }
}