using System;
using System.Collections.Generic;
using System.Linq;
using System.Collections.ObjectModel;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

namespace AvatarHoldingObject
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        AvatarDescription avatarDescription;
        AvatarAnimation avatarAnimation;
        AvatarRenderer avatarRenderer;

        Matrix world;
        Matrix view;
        Matrix projection;

        GamePadState lastGamePadState;

        // Avatar bones in world space
        List<Matrix> bonesWorldSpace;

        Model handWeight;
        Matrix handWorldMatrix;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;
            Content.RootDirectory = "Content";

            Components.Add(new GamerServicesComponent(this));
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // Load random avatar
            avatarDescription = AvatarDescription.CreateRandom();
            // Load celebrate animation
            avatarAnimation = new AvatarAnimation(AvatarAnimationPreset.Celebrate);
            // Load avatar renderer for random description
            avatarRenderer = new AvatarRenderer(avatarDescription);

            // Rotate avatar to face the screen
            world = Matrix.CreateRotationY(MathHelper.Pi);
            // Set view and projection matrices to something sensible
            view = Matrix.CreateLookAt(new Vector3(0, 1, 3), new Vector3(0, 1, 0), Vector3.Up);
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4,
                                                             GraphicsDevice.Viewport.AspectRatio,
                                                             0.01f, 200.0f);

            // Load the hand weight model
            handWeight = Content.Load<Model>("HandWeight");

            // Turn on default lighting
            foreach (ModelMesh mesh in handWeight.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                }
            }

            // Populate the world space bone list
            bonesWorldSpace = new List<Matrix>(AvatarRenderer.BoneCount);
            for (int i = 0; i < AvatarRenderer.BoneCount; i++)
                bonesWorldSpace.Add(Matrix.Identity);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // Update the animation with the elapsed frame time
            // Set the looping parameter to true
            avatarAnimation.Update(gameTime.ElapsedGameTime, true);

            if (avatarRenderer.State == AvatarRendererState.Ready)
            {
                ConvertBonesToWorldSpace(avatarRenderer, avatarAnimation);
            }

            // Find the position of the right hand bone
            handWorldMatrix = bonesWorldSpace[(int)AvatarBone.SpecialRight];


            // Get the current state of player one controller
            GamePadState currentGamePadState = GamePad.GetState(PlayerIndex.One);

            // Create new random avatar if the user presses A
            if (currentGamePadState.Buttons.A == ButtonState.Pressed &&
                lastGamePadState.Buttons.A == ButtonState.Released)
            {
                // Generate new random description
                avatarDescription = AvatarDescription.CreateRandom();
                avatarRenderer = new AvatarRenderer(avatarDescription);
            }

            // Save the state so we can prevent repeat button presses
            lastGamePadState = currentGamePadState;

            base.Update(gameTime);
        }

        /// <summary>
        /// Convert the avatar bones into world space
        /// </summary>
        private void ConvertBonesToWorldSpace(AvatarRenderer renderer, AvatarAnimation animation)
        {
            // Bind pose of the avatar skeleton
            // Values are in local space and relative to the parent bone
            IList<Matrix> bindPose = renderer.BindPose;
            // Animation bone transforms are in local space and relative to the parent bone
            IList<Matrix> animationPose = animation.BoneTransforms;
            // Parent index values for each of the bones
            IList<int> parentIndexes = renderer.ParentBones;

            // Transform each bone into world space
            for (int i = 0; i < AvatarRenderer.BoneCount; i++)
            {
                // Partent world space matrix
                // The world matrix of the avatar renderer is used for the root
                Matrix parentMatrix = (parentIndexes[i] != -1)
                                        ? bonesWorldSpace[parentIndexes[i]]
                                        : renderer.World;

                // Calculate world space matrix for the bone
                bonesWorldSpace[i] = Matrix.Multiply(Matrix.Multiply(animationPose[i],
                                                                        bindPose[i]),
                                                                        parentMatrix);
            }
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // Set the world, view, and projection matrices
            avatarRenderer.World = world;
            avatarRenderer.View = view;
            avatarRenderer.Projection = projection;

            // Draw the avatar to the current render target
            avatarRenderer.Draw(avatarAnimation);

            // Draw the hand weight model
            handWeight.Draw(handWorldMatrix, view, projection);

            base.Draw(gameTime);
        }
    }
}
