using System;
using System.Collections.Generic;
using System.Linq;
using System.Collections.ObjectModel;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

namespace MultipleAvatarAnimations
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        AvatarDescription avatarDescription;
        AvatarAnimation avatarAnimationCelebrate;
        AvatarAnimation avatarAnimationWave;
        AvatarRenderer avatarRenderer;

        Matrix world;
        Matrix view;
        Matrix projection;

        GamePadState lastGamePadState;

        // Combined transforms for the two animations
        List<Matrix> combinedBoneTransfroms = new List<Matrix>(AvatarRenderer.BoneCount);

        // Bone index values for the right arm
        List<int> rightArmBones;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;
            Content.RootDirectory = "Content";

            Components.Add(new GamerServicesComponent(this));
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // Load random avatar
            avatarDescription = AvatarDescription.CreateRandom();
            
            // Load celebrate animation
            avatarAnimationCelebrate = new AvatarAnimation(AvatarAnimationPreset.Celebrate);
            // Load the wave animation
            avatarAnimationWave = new AvatarAnimation(AvatarAnimationPreset.Wave);

            // Load avatar renderer for random description
            avatarRenderer = new AvatarRenderer(avatarDescription);

            // Rotate avatar to face the screen
            world = Matrix.CreateRotationY(MathHelper.Pi);
            // Set view and projection matrices to something sensible
            view = Matrix.CreateLookAt(new Vector3(0, 1, 3), new Vector3(0, 1, 0), Vector3.Up);
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4,
                                                             GraphicsDevice.Viewport.AspectRatio,
                                                             0.01f, 200.0f);

            // Find all of the children of the right shoulder
            rightArmBones = FindChildrenBones(AvatarBone.ShoulderRight, avatarRenderer.ParentBones);

            // Populate the list of bone transforms
            for (int i = 0; i < AvatarRenderer.BoneCount; ++i)
            {
                combinedBoneTransfroms.Add(Matrix.Identity);
            }
        }

        private List<int> FindChildrenBones(AvatarBone parentBone, ReadOnlyCollection<int> parentBones)
        {
            List<int> children = new List<int>();
            children.Add((int)parentBone);

            // Start search one past the parent
            int currentBone = (int)parentBone + 1;

            // Loop over all of the bones
            while (currentBone < parentBones.Count)
            {
                // See if the bone has a parent in the list
                if (children.Contains(parentBones[currentBone]))
                {
                    children.Add(currentBone);
                }
                currentBone++;
            }

            return children;
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // Update both animations
            avatarAnimationCelebrate.Update(gameTime.ElapsedGameTime, true);
            avatarAnimationWave.Update(gameTime.ElapsedGameTime, true);

            // Get the current state of player one controller
            GamePadState currentGamePadState = GamePad.GetState(PlayerIndex.One);

            // Create new random avatar if the user presses A
            if (currentGamePadState.Buttons.A == ButtonState.Pressed &&
                lastGamePadState.Buttons.A == ButtonState.Released)
            {
                // Generate new random description
                avatarDescription = AvatarDescription.CreateRandom();
                avatarRenderer = new AvatarRenderer(avatarDescription);
            }

            // Save the state so we can prevent repeat button presses
            lastGamePadState = currentGamePadState;

            // Copy the celebrate transforms
            for (int i = 0; i < combinedBoneTransfroms.Count; i++)
            {
                combinedBoneTransfroms[i] = avatarAnimationCelebrate.BoneTransforms[i];
            }

            // Overwrite the transforms for the right arm using the values 
            // from the wave animation
            for (int i = 0; i < rightArmBones.Count; i++)
            {
                combinedBoneTransfroms[rightArmBones[i]] = avatarAnimationWave.BoneTransforms[rightArmBones[i]];
            }

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // Set the world, view, and projection matrices
            avatarRenderer.World = world;
            avatarRenderer.View = view;
            avatarRenderer.Projection = projection;

            // Draw the avatar using the combined bone transforms
            avatarRenderer.Draw(combinedBoneTransfroms, avatarAnimationCelebrate.Expression);

            base.Draw(gameTime);
        }
    }
}
