using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;

namespace SimpleNetworking
{
    // Different states the game can be in
    public enum GameState { MainMenu, CreateSession, FindSession, GameLobby, PlayingGame };

    // Properties that define what type of game session to create
    public enum SessionProperties { GameType, MapLevel, OtherCustomProperty };
    public enum GameType { Deathmatch, CaptureTheFlag, FreeForAll };

    // Structure used to display game messages over a TimeSpan
    public struct DisplayMessage
    {
        public string Message;
        public TimeSpan DisplayTime;

        public DisplayMessage(string message, TimeSpan displayTime)
        {
            Message = message;
            DisplayTime = displayTime;
        }
    }

    // Each gamer in the session will have their own GameObject
    public class GameObject
    {
        public Vector2 Position
        { get; set; }

        public GameObject(Vector2 position)
        {
            Position = position;
        }
    }

    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        // Current state our game is in we start at the main menu
        GameState gameState = GameState.MainMenu;

        // The network session for the game
        NetworkSession networkSession;

        // List of sessions that you can join
        AvailableNetworkSessionCollection availableSessions;

        // PacketWriter and PacketReader used to send and recieve game data
        PacketWriter packetWriter = new PacketWriter();
        PacketReader packetReader = new PacketReader();

        // The messages to display on the screen
        // We will use this to show events that occur like
        // a player joining a session
        List<DisplayMessage> gameMessages = new List<DisplayMessage>();

        SpriteFont spriteFont;
        GamePadState currentGamePadState;
        GamePadState lastGamePadState;
        Random random = new Random();

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferWidth = 1280;
            graphics.PreferredBackBufferHeight = 720;
            Content.RootDirectory = "Content";

            Components.Add(new GamerServicesComponent(this));

            // Register for the invite event
            NetworkSession.InviteAccepted += new EventHandler<InviteAcceptedEventArgs>(NetworkSession_InviteAccepted);
        }

        void NetworkSession_InviteAccepted(object sender, InviteAcceptedEventArgs e)
        {
            if (networkSession != null && !networkSession.IsDisposed)
                networkSession.Dispose(); 

            if (!e.IsCurrentSession)
                networkSession = NetworkSession.JoinInvited(1);
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            spriteFont = Content.Load<SpriteFont>("SpriteFont1");
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // If there is no user signed in after 5 seconds 
            // then we show the sign in
            if (gameTime.TotalGameTime.Seconds > 5)
            {
                if (Gamer.SignedInGamers.Count == 0 && 
                    !Guide.IsVisible)
                    Guide.ShowSignIn(1, false);
            }

            // Store the current GamePadState
            currentGamePadState = GamePad.GetState(PlayerIndex.One);

            // Switch to determine which update method to call
            // based on the current game state
            switch (gameState)
            {
                case GameState.MainMenu:
                    MainMenuUpdate();
                    break;
                case GameState.CreateSession:
                    CreateSessionUpdate();
                    break;
                case GameState.FindSession:
                    FindSessionUpdate();
                    break;
                case GameState.GameLobby:
                    GameLobbyUpdate();
                    break;
                case GameState.PlayingGame:
                    PlayingGameUpdate(gameTime);
                    break;
            }

            // Store the game pad state for next frame
            lastGamePadState = currentGamePadState;

            // Update the network session we need to check to 
            // see if it is disposed since calling Update on 
            // a disposed NetworkSession will throw an exception
            if (networkSession != null && !networkSession.IsDisposed)
                networkSession.Update();

            // Update the DisplayTime of the current display message
            if (gameMessages.Count > 0)
            {
                DisplayMessage currentMessage = gameMessages[0];
                currentMessage.DisplayTime -= gameTime.ElapsedGameTime;

                // Remove the message if the time is up
                if (currentMessage.DisplayTime <= TimeSpan.Zero)
                {
                    gameMessages.RemoveAt(0);
                }
                else
                {
                    gameMessages[0] = currentMessage;
                }
            }

            base.Update(gameTime);
        }

        // Helper to determine if a button was pressed but will 
        // not allow repeat presses over several frames
        bool ButtonPressed(Buttons button)
        {
            // Don't process buttons when the guide is visable
            if (Guide.IsVisible)
                return false;
            
            return currentGamePadState.IsButtonDown(button) &&
                    lastGamePadState.IsButtonUp(button);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // Switch to call the correct draw method for the 
            // current game state
            switch (gameState)
            {
                case GameState.MainMenu:
                    MainMenuDraw();
                    break;
                case GameState.CreateSession:
                    CreateSessionDraw();
                    break;
                case GameState.FindSession:
                    FindSessionDraw();
                    break;
                case GameState.GameLobby:
                    GameLobbyDraw();
                    break;
                case GameState.PlayingGame:
                    PlayingGameDraw();
                    break;
            }

            // Draw the current display message
            if (gameMessages.Count > 0)
            {
                DisplayMessage currentMessage = gameMessages[0];
                spriteBatch.Begin();
                Vector2 stringSize = spriteFont.MeasureString(gameMessages[0].Message);
                spriteBatch.DrawString(spriteFont, gameMessages[0].Message, new Vector2((1280 - stringSize.X) / 2.0f, 500), Color.White);
                spriteBatch.End();
            }

            base.Draw(gameTime);
        }

        // Update method for the MainMenu GameState
        private void MainMenuUpdate()
        {
            // Create a new session
            if (ButtonPressed(Buttons.A))
                gameState = GameState.CreateSession;
            // Find a session
            else if (ButtonPressed(Buttons.B))
                FindSession();
            // Exit the game
            else if (ButtonPressed(Buttons.Back))
                Exit();
            // Show a party sessions the player can join
            else if (ButtonPressed(Buttons.Y) && !Guide.IsVisible)
                Guide.ShowPartySessions(PlayerIndex.One);
        }

        // Draw method for the MainMenu GameState
        private void MainMenuDraw()
        {
            spriteBatch.Begin();
            spriteBatch.DrawString(spriteFont, "MAIN MENU", new Vector2(10, 10), Color.White);
            spriteBatch.DrawString(spriteFont, "Create Session - Press A", new Vector2(10, 50), Color.White);
            spriteBatch.DrawString(spriteFont, "Find Session - Press B", new Vector2(10, 90), Color.White);
            spriteBatch.DrawString(spriteFont, "Exit - Press Back", new Vector2(10, 130), Color.White);
            spriteBatch.End();
        }

        // Update method for the create session method
        private void CreateSessionUpdate()
        {
            // Move back to the main menu
            if (ButtonPressed(Buttons.Back))
                gameState = GameState.MainMenu;
            // Create a new session with different 
            // using different values for GameType
            else if (ButtonPressed(Buttons.A))
                CreateSession(GameType.Deathmatch);
            else if (ButtonPressed(Buttons.B))
                CreateSession(GameType.CaptureTheFlag);
            else if (ButtonPressed(Buttons.X))
                CreateSession(GameType.FreeForAll);
        }

        // Createa a new NetworkSession
        private void CreateSession(GameType gameType)
        {
            // If we have an existing network session we need to dispose of it
            if (networkSession != null && !networkSession.IsDisposed)
                networkSession.Dispose();

            // Create the NetworkSessionProperties to use for the session
            // Other players will use these to search for a session
            NetworkSessionProperties sessionProperties = new NetworkSessionProperties();
            sessionProperties[(int)SessionProperties.GameType] = (int)gameType;
            sessionProperties[(int)SessionProperties.MapLevel] = 0;
            sessionProperties[(int)SessionProperties.OtherCustomProperty] = 42;

            // Create the NetworkSession NetworkSessionType of SystemLink
            networkSession = NetworkSession.Create(NetworkSessionType.SystemLink, 1, 4, 0, sessionProperties);
            networkSession.AllowJoinInProgress = true;

            // Register for NetworkSession events
            networkSession.GameStarted += new EventHandler<GameStartedEventArgs>(networkSession_GameStarted);
            networkSession.GameEnded += new EventHandler<GameEndedEventArgs>(networkSession_GameEnded);
            networkSession.GamerJoined += new EventHandler<GamerJoinedEventArgs>(networkSession_GamerJoined);
            networkSession.GamerLeft += new EventHandler<GamerLeftEventArgs>(networkSession_GamerLeft);
            networkSession.SessionEnded += new EventHandler<NetworkSessionEndedEventArgs>(networkSession_SessionEnded);

            // Move the game into the GameLobby state
            gameState = GameState.GameLobby;
        }

        // Event handler for the NetworkSession.GameStarted event
        // This event is fired when the host call NetworkSession.StartGame
        void networkSession_GameStarted(object sender, GameStartedEventArgs e)
        {
            gameMessages.Add(new DisplayMessage("Game Started", TimeSpan.FromSeconds(2)));
            // Move the game into the PlayingGame state
            gameState = GameState.PlayingGame;
        }

        // Event handler for the NetworkSession.GameEnded event
        // This event is fired when the host call NetworkSession.EndGame
        void networkSession_GameEnded(object sender, GameEndedEventArgs e)
        {
            gameMessages.Add(new DisplayMessage("Game Ended", TimeSpan.FromSeconds(2)));
            // Move the game into the GameLobby state
            gameState = GameState.GameLobby;
        }

        // Event handler for the NetworkSession.GamerJoined event
        // This event is fired when someone joins the session
        // This event will fire even for local gamers
        void networkSession_GamerJoined(object sender, GamerJoinedEventArgs e)
        {
            gameMessages.Add(new DisplayMessage("Gamer Joined: " + e.Gamer.Gamertag, TimeSpan.FromSeconds(2)));
            // Add a new GameObject that we will use to store game state for the player
            e.Gamer.Tag = new GameObject(new Vector2(random.Next(100, 1000), random.Next(100, 600)));
        }

        // Event handler for the NetworkSession.GamerLeft event
        // This event is fired when a player leaves the session
        void networkSession_GamerLeft(object sender, GamerLeftEventArgs e)
        {
            gameMessages.Add(new DisplayMessage("Gameer Left: " + e.Gamer.Gamertag, TimeSpan.FromSeconds(2)));
        }

        // Event handler for the NetworkSession.SessionEnded event
        // This event is fired when your connection to the NetworkSession is ended
        void networkSession_SessionEnded(object sender, NetworkSessionEndedEventArgs e)
        {
            gameMessages.Add(new DisplayMessage("Session Ended: " + e.EndReason.ToString(), TimeSpan.FromSeconds(2)));

            // Since we have disconnected we clean up the NetworkSession
            if (networkSession != null && !networkSession.IsDisposed)
                networkSession.Dispose();

            // Move the game into the MainMenu state
            gameState = GameState.MainMenu;
        }

        // Draw method for the CreateSession GameState
        private void CreateSessionDraw()
        {
            spriteBatch.Begin();
            spriteBatch.DrawString(spriteFont, "CREATE SESSION", new Vector2(10, 10), Color.White);
            spriteBatch.DrawString(spriteFont, "Deathmatch - Press A", new Vector2(10, 50), Color.White);
            spriteBatch.DrawString(spriteFont, "CaptureTheFlag - Press B", new Vector2(10, 90), Color.White);
            spriteBatch.DrawString(spriteFont, "FreeForAll - Press X", new Vector2(10, 130), Color.White);
            spriteBatch.DrawString(spriteFont, "Exit - Press Back", new Vector2(10, 170), Color.White);
            spriteBatch.End();
        }

        // Update method for the GameLobby GameState 
        private void GameLobbyUpdate()
        {
            // Move back to the main menu
            if (ButtonPressed(Buttons.Back))
            {
                networkSession.Dispose();
                gameState = GameState.MainMenu;
            }
            // Set the ready state for the player
            else if (ButtonPressed(Buttons.A))
                networkSession.LocalGamers[0].IsReady = !networkSession.LocalGamers[0].IsReady;
            // Only the host can start the game
            else if (ButtonPressed(Buttons.Start) && networkSession.IsHost)
                networkSession.StartGame();
            // Invite other players
            else if (ButtonPressed(Buttons.X) && !Guide.IsVisible)
                Guide.ShowGameInvite(PlayerIndex.One, null);
            // Show the party screen
            else if (ButtonPressed(Buttons.Y) && !Guide.IsVisible)
                Guide.ShowParty(PlayerIndex.One);
        }

        // Draw method for the GameLobby GameState
        private void GameLobbyDraw()
        {
            spriteBatch.Begin();
            spriteBatch.DrawString(spriteFont, "GAME LOBBY", new Vector2(10, 10), Color.White);
            spriteBatch.DrawString(spriteFont, "Mark Ready Status - Press A", new Vector2(10, 50), Color.White);            
            spriteBatch.DrawString(spriteFont, "Exit - Press Back", new Vector2(10, 90), Color.White);
            if (networkSession.IsHost)
                spriteBatch.DrawString(spriteFont, "Start Game - Press Start", new Vector2(10, 130), Color.White);

            // Draw all games in the lobby
            spriteBatch.DrawString(spriteFont, "PLAYERS IN LOBBY", new Vector2(10, 220), Color.White);
            float drawOffset = 0;
            foreach (NetworkGamer networkGamer in networkSession.AllGamers)
            {
                spriteBatch.DrawString(spriteFont, networkGamer.Gamertag + " - " + 
                                                    ((networkGamer.IsReady) ? "READY" : "NOT READY") + " " +
                                                    ((networkGamer.IsTalking) ? "TALKING" : ""), 
                                                    new Vector2(10, 260 + drawOffset), Color.White);
                drawOffset += 40;
            }

            spriteBatch.End();
        }

        // Update method for the PlayingGame GameState
        private void PlayingGameUpdate(GameTime gameTime)
        {
            // Check to see if the player wants to quit
            if (ButtonPressed(Buttons.Back))
            {
                // If the player is the host then 
                // the game is exited but the session
                // stays alive
                if (networkSession.IsHost)
                    networkSession.EndGame();
                // Other players leave the session 
                else
                {
                    networkSession.Dispose();
                    networkSession = null;
                    gameState = GameState.MainMenu;
                }

                return;
            }
                

            // Use GamePad to update the position of the local gamers
            // Loop all of the local gamers
            foreach (LocalNetworkGamer gamer in networkSession.LocalGamers)
            {
                // Handle local input
                GamePadState gamePadState = GamePad.GetState(gamer.SignedInGamer.PlayerIndex);

                // Get the GameObject for this local gamer
                GameObject gameObject = gamer.Tag as GameObject;
                Vector2 position = gameObject.Position;

                // Update the position of the game object based on the GamePad
                Vector2 move = gamePadState.ThumbSticks.Left * (float)gameTime.ElapsedGameTime.TotalSeconds * 500;
                move.Y *= -1;
                position += move;
                gameObject.Position = position;

                // Write the new position to the packet writer
                packetWriter.Write(position);

                // Send the data to everyone in your session
                gamer.SendData(packetWriter, SendDataOptions.InOrder);
            }

            // Read data that is sent to local players
            foreach (LocalNetworkGamer gamer in networkSession.LocalGamers)
            {
                // Read until there is no data left
                while (gamer.IsDataAvailable)
                {
                    NetworkGamer sender;
                    gamer.ReceiveData(packetReader, out sender);

                    // We only need to update the state of non local players
                    if (sender.IsLocal)
                        continue;

                    // Get GameObject of the sender
                    GameObject gameObject = sender.Tag as GameObject;

                    // Read the position
                    gameObject.Position = packetReader.ReadVector2();
                }
            }
        }

        // Draw method for the PlayingGame GameState
        private void PlayingGameDraw()
        {
            spriteBatch.Begin();
            spriteBatch.DrawString(spriteFont, "PLAYING GAME", new Vector2(10, 10), Color.White);

            // Draw each players name at their position in the game
            foreach (NetworkGamer networkGamer in networkSession.AllGamers)
            {
                GameObject player = networkGamer.Tag as GameObject;

                spriteBatch.DrawString(spriteFont, networkGamer.Gamertag +
                                                    ((networkGamer.IsHost) ? " (HOST)": "") +
                                                    ((networkGamer.IsTalking) ? " TALKING" : ""),
                                                    player.Position, 
                                                    Color.White);
            }

            spriteBatch.End();
        }

        // Method to start the search for a NetworkSession
        private void FindSession()
        {
            // Dispose of any previous session
            if (networkSession != null && !networkSession.IsDisposed)
                networkSession.Dispose();

            // Define the type of session we want to search for using the NetworkSessionProperties. 
            // We only set the MapLevel and the OtherCustomProperty
            NetworkSessionProperties sessionProperties = new NetworkSessionProperties();
            sessionProperties[(int)SessionProperties.MapLevel] = 0;
            sessionProperties[(int)SessionProperties.OtherCustomProperty] = 42;

            // Find an available NetworkSession
            availableSessions = NetworkSession.Find(NetworkSessionType.SystemLink, 1, sessionProperties);

            // Move the game into the FindSession state 
            gameState = GameState.FindSession;
        }

        // Update method for the FindSession GameState
        private void FindSessionUpdate()
        {
            // Go back to the main menu
            if (ButtonPressed(Buttons.Back))
                gameState = GameState.MainMenu;
            // If the user presses the A button join the first session
            else if (ButtonPressed(Buttons.A) && availableSessions.Count != 0)
                JoinSession(0);
        }

        // Draw method for the FindSession GameState
        private void FindSessionDraw()
        {
            spriteBatch.Begin();
            spriteBatch.DrawString(spriteFont, "FIND SESSION", new Vector2(10, 10), Color.White);
            spriteBatch.DrawString(spriteFont, "Exit - Press Back", new Vector2(10, 50), Color.White);

            // Write message if there are no sessions found
            if (availableSessions.Count == 0)
                spriteBatch.DrawString(spriteFont, "NO SESSIONS FOUND", new Vector2(10, 90), Color.White);
            else
            {
                // Print out a list of the available sessions
                int sessionIndex = 0;
                foreach (AvailableNetworkSession session in availableSessions)
                {
                    spriteBatch.DrawString(spriteFont, session.HostGamertag + " " +
                                                        session.OpenPublicGamerSlots +
                                                        ((sessionIndex == 0) ? " (PRESS A)" : ""),
                                                        new Vector2(10, 90 + sessionIndex * 40), Color.White);
                    sessionIndex++;
                }
            }

            spriteBatch.End();
        }

        // Creates a session using a AvailableNetworkSession index
        private void JoinSession(int sessionID)
        {
            // Join an existing NetworkSession
            try
            {
                networkSession = NetworkSession.Join(availableSessions[sessionID]);
            }
            catch (NetworkSessionJoinException ex)
            {
                gameMessages.Add(new DisplayMessage("Faild to connect to session: " +
                                                    ex.JoinError.ToString(), 
                                                    TimeSpan.FromSeconds(2)));

                // Check for sessons again
                FindSession();
            }

            // Register for NetworkSession events
            networkSession.GameStarted += new EventHandler<GameStartedEventArgs>(networkSession_GameStarted);
            networkSession.GameEnded += new EventHandler<GameEndedEventArgs>(networkSession_GameEnded);
            networkSession.GamerJoined += new EventHandler<GamerJoinedEventArgs>(networkSession_GamerJoined);
            networkSession.GamerLeft += new EventHandler<GamerLeftEventArgs>(networkSession_GamerLeft);
            networkSession.SessionEnded += new EventHandler<NetworkSessionEndedEventArgs>(networkSession_SessionEnded);

            // Set the correct GameState. The NetworkSession may have already started a game.
            if (networkSession.SessionState == NetworkSessionState.Playing)
                gameState = GameState.PlayingGame;
            else
                gameState = GameState.GameLobby;
        }
    }
}
